/*
	"Multi-Tool"

	ALPHA v0.5 (not for general release)

			MANUAL;
		Attack-Release enveloppe / LFO multimode module with tap-tempo option.

		A-R Enveloppe mode (default)
		As an enveloppe you set the total duration with the top knob and left
		mod input. Higher values will lead to a shorter total duration. The 2nd
		knob and right modulation input will affect the ratio betweeen attack
		and decay. Higher values will lead to a more percussive style of
		modulation, making this suitable for using keyboard velocity as a
		modulation source. Trig input or button A (re)triggers the enveloppe.
		Toggle between LFO and Envelope mode using button B. In this mode led B
		is off.

		LFO mode
		As an LFO the module works more or less the same, except it loops and
		does not need to be triggered. This means the first modulation source
		sets the rate, while the second morphs gradulally between an upward saw
		and a downward one, with a triangle shape in the middle. In this mode
		the trigger input is not needed, though it may be used to reset the
		LFO. Button A manually resets the LFO. In this mode led B is on.

		Tap-Tempo LFO
		In this mode the LFO's rate is set based on tapping button A or
		supplying a sync signal on the trig input. Knob 1 and the related
		modulation input now sets a multiplyer on this tempo, ranging from
		a quarter of the tempo you tapped out to 8 times this rate. For a one
		on one rate set knob 1 to about 11 o'clock. Knob and input 2 function
		as they do in plain LFO mode. Hold button A, then press B, to access
		this mode. In this mode led B blinks.

		In all modes led A indicates the signal on the trig input.
 */
 
#include "SPI.h" // spi library
#include <avr/interrupt.h> //library for using interupts
#include <avr/pgmspace.h>
#include "kas-button.h" //button debouncing

static byte mod_a_pin = A1;
static byte mod_b_pin = A2;
static byte trigger_pin = 2;
//static byte trigger_pin = 4;
static byte but_A_pin = 1;
static byte but_B_pin = 9;
static byte led_a = 3;
static byte led_b = 5;

static unsigned long max_u_long = 4294967295;
static unsigned long max_u_int = 65535;

unsigned phase = 0;
unsigned break_point = 256;
unsigned phase_inc = 1;
unsigned tt_phase = 0;
unsigned measured_phase = 0;

word max_count = 1;
word trig_count = 0;

unsigned long out_val = 0;
unsigned long up_step = 1;
unsigned long down_step = 1;

word mod_a_val = 0;
word mod_b_val = 0;
Button trigger;
Button but_A;
Button but_B;
bool trigger_state;
bool trigger_button_state;
bool button_state;
bool looping = false;
bool playing = false;
bool tt_mode = false;


//exponential table ranging from 4 seconds to 20ms
const uint16_t freqTable[] PROGMEM = {
16, 16, 16, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 18, 18, 18,
18, 18, 18, 18, 18, 18, 18, 18, 18, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19,
19, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 21, 21, 21, 21, 21, 21, 21,
21, 21, 21, 21, 22, 22, 22, 22, 22, 22, 22, 22, 22, 22, 23, 23, 23, 23, 23, 23,
23, 23, 23, 23, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 25, 25, 25, 25, 25, 25,
25, 25, 25, 26, 26, 26, 26, 26, 26, 26, 26, 26, 27, 27, 27, 27, 27, 27, 27, 27,
28, 28, 28, 28, 28, 28, 28, 28, 29, 29, 29, 29, 29, 29, 29, 29, 30, 30, 30, 30,
30, 30, 30, 30, 31, 31, 31, 31, 31, 31, 31, 32, 32, 32, 32, 32, 32, 32, 33, 33,
33, 33, 33, 33, 33, 34, 34, 34, 34, 34, 34, 34, 35, 35, 35, 35, 35, 35, 35, 36,
36, 36, 36, 36, 36, 37, 37, 37, 37, 37, 37, 38, 38, 38, 38, 38, 38, 39, 39, 39,
39, 39, 39, 40, 40, 40, 40, 40, 40, 41, 41, 41, 41, 41, 41, 42, 42, 42, 42, 42,
43, 43, 43, 43, 43, 43, 44, 44, 44, 44, 44, 45, 45, 45, 45, 45, 46, 46, 46, 46,
46, 47, 47, 47, 47, 47, 48, 48, 48, 48, 48, 49, 49, 49, 49, 49, 50, 50, 50, 50,
51, 51, 51, 51, 51, 52, 52, 52, 52, 53, 53, 53, 53, 53, 54, 54, 54, 54, 55, 55,
55, 55, 56, 56, 56, 56, 57, 57, 57, 57, 58, 58, 58, 58, 59, 59, 59, 59, 60, 60,
60, 60, 61, 61, 61, 61, 62, 62, 62, 62, 63, 63, 63, 64, 64, 64, 64, 65, 65, 65,
66, 66, 66, 66, 67, 67, 67, 68, 68, 68, 68, 69, 69, 69, 70, 70, 70, 71, 71, 71,
71, 72, 72, 72, 73, 73, 73, 74, 74, 74, 75, 75, 75, 76, 76, 76, 77, 77, 77, 78,
78, 78, 79, 79, 79, 80, 80, 80, 81, 81, 81, 82, 82, 82, 83, 83, 83, 84, 84, 84,
85, 85, 86, 86, 86, 87, 87, 87, 88, 88, 89, 89, 89, 90, 90, 91, 91, 91, 92, 92,
92, 93, 93, 94, 94, 94, 95, 95, 96, 96, 97, 97, 97, 98, 98, 99, 99, 99, 100,
100, 101, 101, 102, 102, 103, 103, 103, 104, 104, 105, 105, 106, 106, 107, 107,
107, 108, 108, 109, 109, 110, 110, 111, 111, 112, 112, 113, 113, 114, 114, 115,
115, 116, 116, 117, 117, 118, 118, 119, 119, 120, 120, 121, 121, 122, 122, 123,
123, 124, 124, 125, 125, 126, 127, 127, 128, 128, 129, 129, 130, 130, 131, 132,
132, 133, 133, 134, 134, 135, 136, 136, 137, 137, 138, 139, 139, 140, 140, 141,
142, 142, 143, 143, 144, 145, 145, 146, 146, 147, 148, 148, 149, 150, 150, 151,
152, 152, 153, 154, 154, 155, 156, 156, 157, 158, 158, 159, 160, 160, 161, 162,
162, 163, 164, 164, 165, 166, 167, 167, 168, 169, 169, 170, 171, 172, 172, 173,
174, 175, 175, 176, 177, 178, 178, 179, 180, 181, 182, 182, 183, 184, 185, 185,
186, 187, 188, 189, 189, 190, 191, 192, 193, 194, 194, 195, 196, 197, 198, 199,
200, 200, 201, 202, 203, 204, 205, 206, 206, 207, 208, 209, 210, 211, 212, 213,
214, 215, 216, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228,
229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244,
245, 246, 247, 248, 249, 250, 251, 253, 254, 255, 256, 257, 258, 259, 260, 261,
263, 264, 265, 266, 267, 268, 269, 271, 272, 273, 274, 275, 276, 278, 279, 280,
281, 282, 284, 285, 286, 287, 288, 290, 291, 292, 293, 295, 296, 297, 299, 300,
301, 302, 304, 305, 306, 308, 309, 310, 312, 313, 314, 316, 317, 318, 320, 321,
322, 324, 325, 327, 328, 329, 331, 332, 334, 335, 337, 338, 340, 341, 342, 344,
345, 347, 348, 350, 351, 353, 354, 356, 357, 359, 360, 362, 364, 365, 367, 368,
370, 371, 373, 375, 376, 378, 380, 381, 383, 384, 386, 388, 389, 391, 393, 394,
396, 398, 400, 401, 403, 405, 406, 408, 410, 412, 413, 415, 417, 419, 421, 422,
424, 426, 428, 430, 432, 433, 435, 437, 439, 441, 443, 445, 447, 449, 450, 452,
454, 456, 458, 460, 462, 464, 466, 468, 470, 472, 474, 476, 478, 480, 482, 484,
487, 489, 491, 493, 495, 497, 499, 501, 504, 506, 508, 510, 512, 514, 517, 519,
521, 523, 526, 528, 530, 532, 535, 537, 539, 542, 544, 546, 549, 551, 553, 556,
558, 560, 563, 565, 568, 570, 573, 575, 578, 580, 582, 585, 587, 590, 593, 595,
598, 600, 603, 605, 608, 611, 613, 616, 618, 621, 624, 626, 629, 632, 635, 637,
640, 643, 646, 648, 651, 654, 657, 660, 662, 665, 668, 671, 674, 677, 680, 682,
685, 688, 691, 694, 697, 700, 703, 706, 709, 712, 715, 718, 722, 725, 728, 731,
734, 737, 740, 744, 747, 750, 753, 756, 760, 763, 766, 769, 773, 776, 779, 783,
786, 789, 793, 796, 800, 803, 807, 810, 813, 817, 820, 824, 828, 831, 835, 838,
842, 845, 849, 853, 856, 860, 864, 867, 871, 875, 879, 882, 886, 890, 894, 898,
902, 905, 909, 913, 917, 921, 925, 929, 933, 937, 941, 945, 949, 953, 957, 961,
965, 970, 974, 978, 982, 986, 991, 995, 999, 1003, 1008, 1012, 1016, 1021,
1025, 1029, 1034, 1038, 1043, 1047, 1052, 1056, 1061, 1065, 1070, 1075, 1079,
1084, 1088, 1093, 1098, 1102, 1107, 1112, 1117, 1122, 1126, 1131, 1136, 1141,
1146, 1151, 1156, 1161, 1166, 1171, 1176, 1181, 1186, 1191, 1196, 1201, 1206,
1211, 1217, 1222, 1227, 1232, 1238, 1243, 1248, 1254, 1259, 1264, 1270, 1275,
1281, 1286, 1292, 1297, 1303, 1308, 1314, 1320, 1325, 1331, 1337
};


uint16_t mapFreq(uint16_t input) {
  return pgm_read_word_near(freqTable + input);
}

void setup()
{
	cli();							// disable global interrupts
	TCCR2A = 0;						// set entire TCCR1A register to 0
	TCCR2B = 0;						// same for TCCR1B
	TCCR2B =  0b00000011;			//scaling to a 256th of the CPU freq = 31250 Hz
	// enable Timer1 overflow interrupt:
	TIMSK2 |= (1 << TOIE1);			//interupt enable
	TCNT2 = 0;						//timer pre-loading
	sei();							//re-enable global interupts

	//set pin(s) to input and output
	pinMode(10, OUTPUT);
	SPI.begin();					// wake up the SPI bus. This is the fast way of controling the DAC
	SPI.setBitOrder(MSBFIRST);
	pinMode(trigger_pin, INPUT_PULLUP);
	pinMode(but_A_pin, INPUT_PULLUP);
	pinMode(but_B_pin, INPUT_PULLUP);
}

//send calculated value to the dac
void out( word value)
{
	byte data;						//spi deals in bytes (8 bits) chunks
	digitalWrite(10, LOW);			//enable writing
	data = highByte(value);			//this includes the 4 most signifficant bits
	data = 0b00001111 & data;		//the rest of the byte is settings for the dac: mask those
	data = 0b00110000 | data;		//set those setting bits (see DAC data sheet or trust me)
	SPI.transfer(data);				//actually send this out
	data = lowByte(value);			//the 8 least signifficant bits of our value
	SPI.transfer(data);				//send them
	digitalWrite(10, HIGH);			//close connection, allowing the DAC to update
}

void setRatio( word r)
{
	r = ~r;
	break_point = r;
	up_step = max_u_long / r;
	down_step = max_u_long / (max_u_int - r);
}

void onTrigger()
{
	if (tt_mode)
	{
		trig_count++;
		measured_phase = max_u_int / tt_phase;
		calcTtBpm();
		if (trig_count >= max_count)
		{
			trig_count = 0;
			phase = 0;
		}
	}
	else
	{
		phase = 0;
	}
	tt_phase = 0;
	playing = true;
}

void calcTtBpm()
{
	switch(mod_a_val >> 7)
	{
		case 0:
			phase_inc  = (measured_phase >> 2);
			max_count = 4;
			break;
		case 1:
			phase_inc  = ((measured_phase >> 2) + (measured_phase >> 1));
			max_count = 3;
			break;
		case 2:
			phase_inc  = (measured_phase >> 1);
			max_count = 2;
			break;
		case 3:
			phase_inc  = measured_phase;
			max_count = 1;
			break;
		case 4:
			phase_inc  = (measured_phase * 2);
			max_count = 1;
			break;
		case 5:
			phase_inc  = (measured_phase * 3);
			max_count = 1;
			break;
		case 6:
			phase_inc  = (measured_phase * 4);
			max_count = 1;
			break;
		case 7:
			phase_inc  = (measured_phase * 8);
			max_count = 1;
			break;
		default:		//shouldn't happen
			phase_inc  = measured_phase;
			max_count = 1;
			break;
	}
}

void loop()
{
	mod_a_val = analogRead(mod_a_pin); //10 bits
	if (tt_mode)
	{
		digitalWrite(led_b, (phase < 2000 ));
	}
	else
	{
		phase_inc = mapFreq(mod_a_val);
	}
	mod_b_val = analogRead(mod_b_pin);
	setRatio(mod_b_val << 6);
	if( updateButton( &trigger, !(digitalRead( trigger_pin ))))
	{
		if (trigger.state)
		{
			onTrigger();
		}
		digitalWrite(led_a, trigger.state);
	}
	if( updateButton( &but_A, !(digitalRead( but_A_pin ))))
	{
		if (but_A.state)
		{
			onTrigger();
		}
		//digitalWrite(led_a, but_A.state);
	}
	if( updateButton( &but_B, !(digitalRead( but_B_pin ))))
	{
		if (but_B.state)
		{
			if (but_A.state)
			{
				looping = true;
				tt_mode = true;
				trig_count = 0;
			}
			else
			{
				looping = !looping;
				if (looping) playing = true;
				tt_mode = false;
			}
		}
		digitalWrite(led_b, looping);
	}
}


ISR(TIMER2_OVF_vect)
{
	tt_phase++;
	if (playing)
	{
		phase += phase_inc;
		if (phase < break_point)
		{
			out_val = phase * up_step;
		}
		else
		{
			out_val = max_u_long - ((phase - break_point) * down_step);
		}
		if (looping) out(out_val >> 20);				//bipolar LFO
		else out((out_val >> 21) + 2048);				//unipolar enveloppe
	}
	if (phase < phase_inc)
	{
		if (!looping)
		{
			playing = false;
			out(2048);
		}
		else if (tt_mode)
		{
			calcTtBpm();
		}
	}
}

